"""This module contains code that can parse one of the plain text NEC question
pool files. Or, well, it can try to. Some extensions are supplied to allow for
marking the file up with extra metadata, like the creation and expiry dates,
the location for graphical figures, a version number, and so forth.
"""

import datetime
import enum
import logging
import re
from io import StringIO
from palmve import pool, utils


logger = logging.getLogger(__name__)


class ParseError(RuntimeError):

    def __init__(self, line_nr, line, error):
        super().__init__('line %d: %s (%s)' % (
            line_nr, error, repr(line)))
        self.line_nr = line_nr
        self.line = line
        self.error = error


def _TitleCase(s):
    """Buggy-ish title casing from vec_sub_element::vec_sub_element"""
    i = StringIO()
    ignore = True
    for c in s:
        if not ignore:
            c = c.lower()
        else:
            ignore = False
        i.write(c)
        if c == ' ':
            ignore = True
    return i.getvalue()


class TextLoader:

    class State(enum.Enum):
        WAIT_FOR_ELEMENT = 1
        IN_ELEMENT = 2
        IN_SUBELEMENT = 3
        IN_GROUP = 4
        IN_QUESTION = 5

    def __init__(self):
        self._element = None
        self._state = self.State.WAIT_FOR_ELEMENT
        self._handlers = {
                self.State.WAIT_FOR_ELEMENT: self._wait_for_element,
                self.State.IN_ELEMENT: self._in_element,
                self.State.IN_SUBELEMENT: self._in_subelement,
                self.State.IN_GROUP: self._in_group,
                self.State.IN_QUESTION: self._in_question,
        }

    def _wait_for_element(self, line_nr, line):
        m = re.match(r'.*ELEMENT\s+(\d+)\s*-\s*(.*)', line)
        if m is not None:
            str_nr, name = m.groups()
            str_nr = int(str_nr)
            self._element = pool.Element(name, str_nr)
            logger.debug('at line %d, found element %s', line_nr, self._element)
            return self.State.IN_ELEMENT

    def _in_element(self, line_nr, line):
        if line.startswith('DATEFIELD'):
            # I have historically made mistakes when I wrote DATEFIELD.
            m = re.match('DATEFIELD (.*) thr(?:u|ough|ouh) (.*)', line)
            g = m.groups()
            self._element.date_created, self._element.date_expires = (
                utils.StrToDate(g[0]), utils.StrToDate(g[1]))
            logger.debug('at line %d, got dates %s and %s',
                line_nr, self._element.date_created, self._element.date_expires)

        elif line.startswith('WITHDRAWN'):
            qid = line.split(' ')[-1]
            self._element.withdrawn_questions.add(qid)
            logger.debug('at line %d, question %s is withdrawn', line_nr, qid)

        elif line.startswith('GRAPHICSDIR'):
            self._element.graphics_directory = line.split(' ')[-1]
            logger.debug('at line %d, graphics dir is %s',
                line_nr, self._element.graphics_directory)

        elif line.startswith('VERSION'):
            self._element.version = int(line.split(' ')[-1])

        elif line.startswith('SUBELEMENT'):
            m = re.match((r'SUBELEMENT\s*(..)\s*[-?]*\s*(.*)\s+'
                r'\[(\d+) (?:[eE]xam )?[qQ]uestions?\s*[-?]*\s*(\d+) [gG]roups?\].*'), line)
            if m is None:
                raise ParseError(line_nr, line, 'subelement declaration does not parse')
            designation, description, exp_g, exp_q = m.groups()
            exp_g = int(exp_g)
            exp_q = int(exp_q)
            if exp_g != exp_q:
                raise ParseError(line_nr, line, 'expectected group and question count do not match')

            if self._element.subelements:
                prev_se = self._element.subelements[-1]
                if len(prev_se.question_groups) != prev_se.expected_group_count:
                    raise ParseError(line_nr, line, 'was expecting %d groups in subelement, but found %d' % (
                        prev_se.expected_group_count,
                        len(prev_se.question_groups)))

            se = pool.SubElement(designation, _TitleCase(description), exp_g)
            self._element.subelements.append(se)
            logger.debug('at line %d, found subelement %s', line_nr, se)
            return self.State.IN_SUBELEMENT

        else:
            logger.debug('ignoring line %d: %s', line_nr, repr(line))

    def _in_subelement(self, line_nr, line):
        s = self._element.subelements[-1]
        if line.startswith(s.designation) and line[3] == ' ':
            m = re.match(r'(...)\s*-*\s*(.*)', line)
            if m is None:
                raise ParseError(line_nr, line, 'group does not parse')
            designation, descr = m.groups()
            g = pool.Group(designation, descr)
            s.question_groups.append(g)
            logger.debug('at line %d, found group %s', line_nr, g)
            return self.State.IN_GROUP

    def _in_group(self, line_nr, line):
        se = self._element.subelements[-1]
        g = se.question_groups[-1]

        if 'DELETED' in line:
            # something weird in data/pools/extra-after-july-01-16.txt.
            return

        elif line.startswith(g.designation):
            # note that some pools have a regulation reference after the
            # question ID, which I have not historically parsed.
            m = re.match(r'(.....)\s*\(([ABCD])\).*', line)
            if m is None:
                raise ParseError(line_nr, line, 'question does not parse')
            designation, answer = m.groups()
            withdrawn = designation in self._element.withdrawn_questions
            q = pool.Question(designation, answer, withdrawn)
            logger.debug('at line %d, found question %s', line_nr, q)
            g.questions.append(q)
            return self.State.IN_QUESTION

        elif line.startswith(se.designation) and line[3] == ' ':
            logger.debug('at line %d, chaining to IN_SUBELEMENT', line_nr)
            return self._in_subelement(line_nr, line)

        elif line.startswith('SUBELEMENT'):
            logger.debug('at line %d, chaining to IN_ELEMENT', line_nr)
            return self._in_element(line_nr, line)

        elif not g.questions:
            g.description += ' ' + line

    def _in_question(self, line_nr, line):
        se = self._element.subelements[-1]
        q = se.question_groups[-1].questions[-1]
        if not q.answers:
            # we haven't seen any answers yet, lines are put of the question text
            if not q.text:
                q.text = line
                return
            elif line[1:3] != '. ':
                q.text += ' ' + line
                return

        if line[1:3] == '. ':
            letter = line[0]
            text = line[3:]
            a = pool.Answer(letter, text)
            q.answers.append(a)

        elif line.startswith('~'):
            fm = re.search(r'[Ff]igure ([^\s,?]+)', q.text)
            if fm is not None:
                f = fm.groups()[0]
                if f.startswith(se.designation[0]):
                    q.figure_designation = f

            if len(q.answers) != 4:
                raise ParseError(line_nr, line, 'question does not have four answers: %s' % repr(q))

            logger.debug('finished parsing question %s' % q)
            return self.State.IN_GROUP

        elif q.answers:
            a = q.answers[-1]
            a.text += ' ' + line

        else:
            raise ParseError(line_nr, line, 'question body is weird')

    def feed_line(self, line_nr, line):
        handler = self._handlers.get(self._state)
        if handler is None:
            raise ParseError(line_nr, line,
                'no state handler found for state %s' % self._state)
        new_state = handler(line_nr, line)
        if new_state is not None:
            logger.debug('transitioning from state %s to state %s at line %d' % (
                self._state.name, new_state.name, line_nr))
            self._state = new_state

    @classmethod
    def from_plain_text(cls, filename):
        # historically, I've preprocessed the text files into plain ascii,
        # removing smart quotes, etc... we could consider automating some of
        # that.
        parser = cls()
        with open(filename, 'rb') as fp:
            line_nr = 1
            while True:
                line = fp.readline()
                if not line:
                    break
                line = line.strip()
                if line:
                    try: line = line.decode('ascii')
                    except Exception as e:
                        raise ParseError(line_nr, line, 'failed to decode as ascii: %s' % str(e))

                    try: parser.feed_line(line_nr, line)
                    except Exception as e:
                        raise ParseError(line_nr, line, str(e))
                line_nr += 1
        return parser._element
