import glob
import json
import logging
import os
import sqlite3
from palmve import utils


logger = logging.getLogger(__name__)


def _LoadSchema(schema_dir=None):
    if schema_dir is None:
        schema_dir = 'src/android/app/room_schemas/net.rten.palmve.data.pool.PoolDatabase'

    schemas = glob.glob(os.path.join(schema_dir, '*.json'))
    schemas.sort()
    if not schemas:
        raise RuntimeError('no schemas found in %s' % schema_dir)
    logging.debug('using schema from %s', schemas[-1])
    with open(schemas[-1]) as fp:
        schema = json.loads(fp.read())

    statements = []
    for e in schema['database']['entities']:
        name = e['tableName']
        sql = e['createSql']
        sql = sql.replace('${TABLE_NAME}', name)
        statements.append(sql)
        for index in e['indices']:
            sql = index['createSql']
            sql = sql.replace('${TABLE_NAME}', name)
            statements.append(sql)

    for v in schema['database']['views']:
        name = v['viewName']
        sql = v['createSql']
        sql = sql.replace('${VIEW_NAME}', name)
        statements.append(sql)

    for q in schema['database']['setupQueries']:
        statements.append(q)

    return statements

_DB_FORMAT = 3


def ElementToDb(e, filename):
    if os.path.isfile(filename):
        raise RuntimeError('file %s already exists' % filename)

    schema = _LoadSchema()
    with sqlite3.connect(filename) as db:
        # First, setup the schema.
        c = db.cursor()
        c.execute('PRAGMA foreign_keys = ON;')
        for entry in schema:
            c.execute(entry)
        db.commit()

        # Look for figures first.
        f_id = 1
        f_map = {None: None}
        if e.graphics_directory is not None:
            gd = os.path.join(os.path.split(filename)[0], 'graphics',
                    e.graphics_directory)
            logging.debug('graphics dir is %s', gd)
            for fig in glob.glob(os.path.join(gd, '*.png')):
                logging.debug('found figure %s', fig)
                designator = os.path.splitext(os.path.basename(fig))[0]
                with open(fig, 'rb') as f:
                    fig_bytes = f.read()

                c.execute('INSERT INTO Figure VALUES (?, ?, ?)', (
                    f_id, designator, fig_bytes))
                f_map[designator] = f_id
                f_id += 1
            db.commit()

        # Now write the element data.
        c.execute('INSERT INTO Element VALUES (?, ?, ?, ?, ?)', (
            e.number, e.name, _DB_FORMAT, e.version,
            '%s through %s' % (
                utils.DateToStr(e.date_created), utils.DateToStr(e.date_expires))))
        se_id = 1
        g_id = 1
        q_id = 1
        a_id = 1
        for se in e.subelements:
            c.execute('INSERT INTO SubElement VALUES (?, ?, ?, ?)', (
                se_id, e.number, se.designation, se.description))
            for g in se.question_groups:
                c.execute('INSERT INTO QuestionGroup VALUES (?, ?, ?, ?)', (
                    g_id, se_id, g.designation[2:], g.description))
                for q in g.questions:
                    c.execute('INSERT INTO Question VALUES (?, ?, ?, ?, ?, ?)', (
                        q_id, g_id, q.designation[3:], q.is_withdrawn,
                        q.correct_answer, f_map[q.figure_designation]))
                    c.execute(('INSERT INTO FTS_Question (docid, description) '
                               'VALUES (?, ?)'), (q_id, q.text))
                    for a in q.answers:
                        c.execute('INSERT INTO Answer VALUES (?, ?, ?)', (
                            a_id, q_id, a.letter))
                        c.execute(('INSERT INTO FTS_Answer (docid, description) '
                                   'VALUES (?, ?)'), (a_id, a.text))
                        a_id += 1
                    q_id += 1
                g_id += 1
            se_id += 1
            db.commit()

        # Lastly, optimize the layout.
        c.execute('VACUUM')
        db.commit()
