package net.rten.palmve.models;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import androidx.arch.core.executor.testing.InstantTaskExecutorRule;
import androidx.lifecycle.Observer;

import net.rten.palmve.TestEnvironment;
import net.rten.palmve.data.pool.Answer;
import net.rten.palmve.data.pool.PoolDatabase;
import net.rten.palmve.data.test.TestDatabase;
import net.rten.palmve.data.test.TestHistory;
import net.rten.palmve.data.test.TestState;
import net.rten.palmve.repository.PoolRepo;
import net.rten.palmve.repository.TestHistoryRepo;
import net.rten.palmve.utilities.LiveDataTestUtil;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;

import javax.inject.Inject;

import dagger.hilt.android.testing.HiltAndroidRule;
import dagger.hilt.android.testing.HiltAndroidTest;

@HiltAndroidTest
public class TestTakingModelTest {
    @Rule
    public InstantTaskExecutorRule _executor = new InstantTaskExecutorRule();
    @Rule
    public final HiltAndroidRule _hiltRule = new HiltAndroidRule(this);
    @Inject
    public TestHistoryRepo _tests;
    @Inject
    public PoolRepo _pools;
    @Inject
    public TestEnvironment _testEnv;

    private TestTakingModel _model = new TestTakingModel();

    @Before
    public void setUp() throws IOException {
        _hiltRule.inject();
        _tests.close();
        _testEnv.reset();
        _tests.reset();
        resetModel();
    }

    @Test
    public void testCreateATestBasics() {
        // setUp() will have started a new test. Make sure the data makes sense.
        Long testId = LiveDataTestUtil.getOrAwaitValue(_model.getTestId());
        TestHistory testInfo = LiveDataTestUtil.getOrAwaitValue(_model.getTestInfo());
        List<TestQuestionData> testQuestions = LiveDataTestUtil.getOrAwaitValue(_model.getTestQuestions());
        Integer currentQuestionIndex = LiveDataTestUtil.getOrAwaitValue(_model.getCurrentQuestionIndex());
        TestTakingModel.TestProgress progress = LiveDataTestUtil.getOrAwaitValue(_model.getTestProgress());

        assertTrue(testId > 0);
        assertEquals(testInfo.testState, TestState.IN_PROGRESS);
        assertEquals(50, testQuestions.size());
        assertEquals(new Integer(0), currentQuestionIndex);
        assertEquals(0, progress.answersCorrect);
        assertEquals(0, progress.percentAnswered);
        assertEquals(0, progress.answersProvided);
        assertEquals(0, progress.percentCorrect);
        assertFalse(progress.allQuestionsAnswered);
    }

    @Test
    public void testAceIt() {
        // We'll answer all questions correctly.
        final List<TestTakingModel.TestProgress> progressUpdates = takeTest(
                (index, data) -> data.testQuestion.correctAnswer);

        // Now take a look at the progress updates. We should have gotten one per question, and
        // the progress values should make sense.
        assertEquals(51, progressUpdates.size());
        for (int i = 0; i < progressUpdates.size(); ++i) {
            assertEquals(i, progressUpdates.get(i).answersProvided);
            assertEquals(i * 100L / (progressUpdates.size() - 1),
                    progressUpdates.get(i).percentAnswered);
            assertEquals(i, progressUpdates.get(i).answersCorrect);
            assertEquals(i > 0 ? 100 : 0, progressUpdates.get(i).percentCorrect);
            if (i == progressUpdates.size() - 1) {
                assertTrue(progressUpdates.get(i).allQuestionsAnswered);
            } else {
                assertFalse(progressUpdates.get(i).allQuestionsAnswered);
            }
        }

        // The current index should have gone back to null.
        assertNull(LiveDataTestUtil.getOrAwaitValue(
                _model.getCurrentQuestionIndex()));

        // We should have passed.
        TestHistory info = _model.getTestInfo().getValue();
        assertNotNull(info);
        assertEquals(TestState.PASSED, info.testState);
        assertEquals(new Integer(100), info.testScorePercent);
    }

    @Test
    public void testFailIt() {
        // We'll answer all questions incorrectly.
        final List<TestTakingModel.TestProgress> progressUpdates = takeTest(
                (index, data) -> incorrectAnswer(data));

        // Now take a look at the progress updates. We should have gotten one per question, and
        // the progress values should make sense.
        assertEquals(51, progressUpdates.size());
        for (int i = 0; i < progressUpdates.size(); ++i) {
            assertEquals(i, progressUpdates.get(i).answersProvided);
            assertEquals(i * 100L / (progressUpdates.size() - 1),
                    progressUpdates.get(i).percentAnswered);
            assertEquals(0, progressUpdates.get(i).answersCorrect);
            assertEquals(0, progressUpdates.get(i).percentCorrect);
            if (i == progressUpdates.size() - 1) {
                assertTrue(progressUpdates.get(i).allQuestionsAnswered);
            } else {
                assertFalse(progressUpdates.get(i).allQuestionsAnswered);
            }
        }

        // The current index should have gone back to null.
        assertNull(LiveDataTestUtil.getOrAwaitValue(
                _model.getCurrentQuestionIndex()));

        // We should have passed.
        TestHistory info = _model.getTestInfo().getValue();
        assertNotNull(info);
        assertEquals(TestState.FAILED, info.testState);
        assertEquals(new Integer(0), info.testScorePercent);
    }

    @Test
    public void testPartial() {
        // Answer some questions and then reset the model.
        Long testId = LiveDataTestUtil.getOrAwaitValue(_model.getTestId());
        final List<TestTakingModel.TestProgress> firstRoundProgressUpdates = takeTest(
                (index, data) -> {
                    if (index >= 0 && index < 10) {
                        return data.testQuestion.correctAnswer;
                    }
                    return null;
                });
        assertEquals(11, firstRoundProgressUpdates.size());
        final Integer firstRoundFinalIndex = LiveDataTestUtil.getOrAwaitValue(
                _model.getCurrentQuestionIndex());
        for (int i = 0; i < firstRoundProgressUpdates.size(); ++i) {
            assertEquals(i, firstRoundProgressUpdates.get(i).answersProvided);
        }
        resetModel();

        // Make sure we get the same test ID back, the test is in progress, etc...
        final Integer secondRoundFirstIndex = LiveDataTestUtil.getOrAwaitValue(
                _model.getCurrentQuestionIndex());
        final TestHistory secondRoundInitialInfo = LiveDataTestUtil.getOrAwaitValue(
                _model.getTestInfo());
        final Long secondRoundTestId = LiveDataTestUtil.getOrAwaitValue(_model.getTestId());
        assertEquals(TestState.IN_PROGRESS, secondRoundInitialInfo.testState);
        assertNull(secondRoundInitialInfo.dateCompleted);
        assertEquals(testId, secondRoundTestId);
        assertEquals(firstRoundFinalIndex, secondRoundFirstIndex);

        // Run through the rest of the questions.
        final List<TestTakingModel.TestProgress> secondRoundProgressUpdates = takeTest(
                (index, data) -> incorrectAnswer(data));
        for (int i = 0; i < secondRoundProgressUpdates.size(); ++i) {
            assertEquals(10 + i, secondRoundProgressUpdates.get(i).answersProvided);
        }
        assertEquals(41, secondRoundProgressUpdates.size());

        // Make sure it was graded as expected.
        final TestHistory secondRoundFinalInfo = LiveDataTestUtil.getOrAwaitValue(
                _model.getTestInfo());
        assertEquals(TestState.FAILED, secondRoundFinalInfo.testState);
        assertEquals(new Integer(10 * 100 / 50), secondRoundFinalInfo.testScorePercent);
    }

    private void resetModel() {
        _model = new TestTakingModel();
        _model.setUp("EXTRA CLASS", "July 1, 2020 throuh June 30, 2024",
                getPoolDb(), getTestDb());
    }

    /**
     * This method will step the model through answering all test questions. If the test completes
     * (all questions answered), it will call completeTest() and return when that's been applied.
     * If TestTaker.pickAnswer() returns null, it will interrupt the test taking process.
     */
    private List<TestTakingModel.TestProgress> takeTest(TestTaker taker) {
        final ArrayList<TestTakingModel.TestProgress> progressUpdates = new ArrayList<>();
        final CountDownLatch latch = new CountDownLatch(1);
        final Observer<Integer> indexChangeObserver = index -> {
            // If the latch has fired, don't do anything. It means we're done.
            if (latch.getCount() == 0) {
                return;
            }

            // Record the progress update if we haven't stopped the test taking process, and if
            // the report is not a duplicate of what we've already recorded. We'll also
            TestTakingModel.TestProgress progress = _model.getTestProgress().getValue();
            if (progress != null && progress.answersProvided > 0 &&
                    (progressUpdates.isEmpty() ||
                            progressUpdates.get(progressUpdates.size() - 1).answersProvided !=
                                    progress.answersProvided)) {
                progressUpdates.add(progress);
            }

            if (index == null) {
                // If index < 0, it means we've run out of questions to answer, and we can complete.
                _model.completeTest();
            } else {
                // If index >= 0, then there is an active question still.
                List<TestQuestionData> questions = _model.getTestQuestions().getValue();
                if (questions != null) {
                    TestQuestionData data = questions.get(index);
                    Answer.AnswerCode code = taker.pickAnswer(index, data);
                    if (code == null) {
                        // Let the test taker tell us that we should abort the process.
                        latch.countDown();
                    } else {
                        _model.submitAnswerToActiveQuestion(code);
                    }
                }
            }
        };
        final Observer<TestHistory> testInfoObserver = info -> {
            if (info.dateCompleted != null && latch.getCount() == 1) {
                latch.countDown();
            }
        };

        // Wait for the first progress update, and put it in the list. Otherwise, we'll have
        // a race condition where sometimes the first '0' progress value will be in the list,
        // sometimes not. This is because duplicate postValue()'s are collapsed.
        TestTakingModel.TestProgress initialProgress = LiveDataTestUtil.getOrAwaitValue(
                _model.getTestProgress());
        progressUpdates.add(initialProgress);

        // Use the observer to step through all questions.
        _model.getCurrentQuestionIndex().observeForever(indexChangeObserver);
        _model.getTestInfo().observeForever(testInfoObserver);
        try {
            if (!latch.await(30, TimeUnit.SECONDS)) {
                throw new RuntimeException("timed out waiting for test completion");
            }
        } catch (InterruptedException e) {
            throw new RuntimeException("await() interrupted");
        } finally {
            _model.getCurrentQuestionIndex().removeObserver(indexChangeObserver);
            _model.getTestInfo().removeObserver(testInfoObserver);
        }

        return progressUpdates;
    }

    private Answer.AnswerCode incorrectAnswer(TestQuestionData question) {
        switch (question.testQuestion.correctAnswer) {
            case A:
                return Answer.AnswerCode.B;
            case B:
                return Answer.AnswerCode.C;
            case C:
                return Answer.AnswerCode.D;
            case D:
                return Answer.AnswerCode.A;
        }
        return null;
    }

    private interface TestTaker {
        Answer.AnswerCode pickAnswer(int questionIndex, TestQuestionData question);
    }

    private TestDatabase getTestDb() {
        return _tests.getDb();
    }

    private PoolDatabase getPoolDb() {
        return _pools.getLocalPools().get(0).getRoomDb();
    }
}
