/*
 * PalmVE - Android amateur radio test practice program
 * Copyright (C) 2010 Michael A Ballbach
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.rten.palmve;

import android.content.Context;
import android.util.Log;

import androidx.annotation.NonNull;
import androidx.room.Room;

import net.rten.palmve.data.pool.Element;
import net.rten.palmve.data.pool.PoolDatabase;

import java.io.File;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;

/**
 * Container for a question pool database.
 */
public class LocalPool implements Pool {
    private static final String[] sMonths;

    static {
        sMonths = new String[]{"January", "February", "March", "April",
                "May", "June", "July", "August", "September", "October",
                "November", "December"};
    }

    private final ArrayList<LocalPoolEvents> _eventListeners;
    private final int _localPoolId;
    private int _elementNr;
    private int _format;
    private int _version;
    private String _name;
    private String _dateField;
    private Status _status;
    private RemotePool _associatedRemotePool;
    private File _fileName;
    private PoolDatabase _roomDb;

    public LocalPool(int localPoolId) {
        _eventListeners = new ArrayList<>();
        _localPoolId = localPoolId;
    }

    /**
     * @return Returns one of Pool's DATE_STATUS_... constants.
     */
    static public int getDateStatusFromDatefield(String datefield) {
        int retv = Pool.DATE_STATUS_ACTIVE;
        EffectiveDateRange edr = getEffectiveDateRange(datefield);

        if (edr.from != null && edr.to != null) {
            Calendar now = Calendar.getInstance();
            if (now.before(edr.from)) {
                retv = Pool.DATE_STATUS_FUTURE;
            } else if (now.after(edr.to)) {
                retv = Pool.DATE_STATUS_EXPIRED;
            }
        }

        return retv;
    }

    /**
     * Parse the range of dates in a date field and get real date objects out of
     * them.
     */
    @NonNull
    static public EffectiveDateRange getEffectiveDateRange(String fromDateField) {
        EffectiveDateRange range = new EffectiveDateRange();

        // okay. unfortunately, I've made mistakes with this field, so we
        // have to be a bit silly about parsing it. what has been consistent
        // is that the string STARTS with a date, and ends with one.
        int offset = -1;
        for (int findSpace = 0; findSpace < 3; findSpace++) {
            offset = fromDateField.indexOf(' ', offset + 1);
            if (offset < 0) {
                break;
            }
        }

        if (offset > 0) {
            String from = fromDateField.substring(0, offset);
            range.from = parseDateRangePart(from);

            // skip to the next space to find the to date
            offset = fromDateField.indexOf(' ', offset + 1);
            if (offset > 0) {
                String to = fromDateField.substring(offset + 1);
                range.to = parseDateRangePart(to);
            }
        }

        return range;
    }

    /**
     * Make a Date() object from the string. I tried using a simple date format
     * object with 'M d, y' as the format string, but it wouldn't parse it. So I
     * do this manually. :/
     */
    static private GregorianCalendar parseDateRangePart(String part) {
        GregorianCalendar retv = null;

        int dayOffset = part.indexOf(' ');
        if (dayOffset > 0 && dayOffset + 1 < part.length()) {
            String month = part.substring(0, dayOffset);
            int yearOffset = part.indexOf(',', dayOffset + 1);
            if (yearOffset > 0 && yearOffset + 2 < part.length()) {
                String day = part.substring(dayOffset + 1, yearOffset);
                String year = part.substring(yearOffset + 2);

                int numericalDay = Integer.parseInt(day);
                int numericalYear = Integer.parseInt(year);
                int numericalMonth = -1;

                for (int t = 0; t < sMonths.length; t++) {
                    if (month.equals(sMonths[t])) {
                        numericalMonth = t;
                        break;
                    }
                }

                if (numericalMonth >= 0) {
                    retv = new GregorianCalendar();
                    retv.clear();
                    retv.set(numericalYear, numericalMonth, numericalDay);
                }
            }
        }

        if (retv == null) {
            Log.w("PalmVE", "Warning: failed to parse datefield part: '" + part
                    + "'");
        }

        return retv;
    }

    /**
     * Experimental access to a Room DB version of the database.
     *
     * @return The room database class for this pool
     */
    public PoolDatabase getRoomDb() {
        return _roomDb;
    }

    /**
     * Add an event handler for pool status changes.
     */
    public void addEventHandler(LocalPoolEvents handler) {
        _eventListeners.add(handler);
    }

    /**
     * Remove an event handler for pool status changes.
     */
    public void removeEventHandler(LocalPoolEvents handler) {
        _eventListeners.remove(handler);
    }

    /**
     * Bind to the pool filename passed, and read the basic pool information.
     * Calling this function should only be done holding the
     * PoolDirectory.POOL_LOCK.
     * <p>
     * Note that this method can't be invoked from the main thread, since it synchronously
     * loads data from a RoomDb.
     */
    public void bindToFile(Context context, File poolFilename) {
        // allow passing a null filename to close/reopen the database
        if (poolFilename == null) {
            poolFilename = _fileName;
        }

        // if we're open, close
        close();

        if (_roomDb == null) {
            // For now, copy it somewhere else. Don't use the same version the CP uses.
            _roomDb = Room.databaseBuilder(context, PoolDatabase.class, poolFilename.toString())
                    .fallbackToDestructiveMigration()
                    .build();
        }

        Element element = _roomDb.elementDao().getElement();
        if (element == null) {
            throw new RuntimeException("expected one and only one element");
        }

        _elementNr = element.id.intValue();  // In the DB it's a long, but it will be 2, 3, or 4.
        _name = element.name;
        _format = element.format;
        _version = element.version;
        _dateField = element.dateField;

        // save the filename
        _fileName = poolFilename;

        // update our status based on our remote pool
        if (_associatedRemotePool != null) {
            setAssociatedRemotePool(_associatedRemotePool);
        } else {
            // assume we're up-to-date unless told otherwise
            setStatus(Status.lpUpToDate);
        }
    }

    /**
     * @return Returns the underlying filename, if there is one.
     */
    public File getUnderlyingFilename() {
        return _fileName;
    }

    /**
     * Close the database. Don't access the pool, except for Pool interface
     * members, after this. Calling this function should only be done holding
     * the PoolDirectory.POOL_LOCK.
     */
    public void close() {
        if (_roomDb != null) {
            _roomDb.close();
            _roomDb = null;
            _fileName = null;
        }
    }

    /**
     * @return Return the associated remote pool, if we have any data about it.
     * Otherwise, returns null.
     */
    public RemotePool getAssociatedRemotePool() {
        return _associatedRemotePool;
    }

    /**
     * Associated the local pool with its remote counterpart. This allows the
     * local pool to determine whether it could be updated from the remote
     * source, or if it is up to date.
     */
    public void setAssociatedRemotePool(RemotePool remotePool) {
        Status newStatus = Status.lpUpToDate;
        _associatedRemotePool = remotePool;
        if (_associatedRemotePool != null && _roomDb != null
                && _associatedRemotePool.getVersion() > getVersion()) {
            newStatus = Status.lpUpdateAvailable;
        }
        setStatus(newStatus);
    }

    /**
     * @return Returns the status of this local pool, whether it could be
     * upgraded, etc...
     */
    public Status getStatus() {
        return _status;
    }

    /**
     * Set the local pool status and call event handlers.
     *
     * @param to New status.
     */
    private void setStatus(Status to) {
        if (to != _status) {
            Status oldStatus = _status;
            _status = to;
            for (LocalPoolEvents handler : _eventListeners) {
                handler.onLocalPoolStatusChanged(this, oldStatus);
            }
        }
    }

    public String getDateField() {
        return _dateField;
    }

    @SuppressWarnings("unused")
    public int getElementNr() {
        return _elementNr;
    }

    public int getFormat() {
        return _format;
    }

    public String getName() {
        return _name;
    }

    public int getVersion() {
        return _version;
    }

    public int getLocalPoolId() {
        return _localPoolId;
    }

    @SuppressWarnings("unused")
    public long getSize() {
        if (_fileName != null) {
            return _fileName.length();
        }
        return 0;
    }

    /**
     * \returns The date this pool becomes effective
     */
    public Calendar getStartEffectiveDate() {
        return getEffectiveDateRange(getDateField()).from;
    }

    /**
     * \returns The date this pool stops being effective
     */
    public Calendar getEndEffectiveDate() {
        return getEffectiveDateRange(getDateField()).to;
    }

    /**
     * @return One of the DATE_STATUS_... constants in the Pool interface,
     * which indicate whether this question pool is active, expired, or
     * a future question pool.
     */
    public int getDateStatus() {
        return getDateStatusFromDatefield(getDateField());
    }

    /**
     * Status of the local pool.
     */
    public enum Status {
        lpUpToDate, lpUpdateAvailable
    }

    /**
     * A range of dates of effectiveness for a question pool.
     */
    static public class EffectiveDateRange {
        public GregorianCalendar from, to;
    }
}
