/*
 * PalmVE - Android amateur radio test practice program
 * Copyright (C) 2010 Michael A Ballbach
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.rten.palmve;

import java.util.ArrayList;
import java.util.Calendar;

/**
 * Represents a remote pool available for downloading.
 *
 * @author ballbach
 */
public class RemotePool implements Pool {
    private final int _elementNr;
    private final int _format;
    private final int _version;
    private final String _name;
    private final String _dateField;
    private final String _sum;
    private final ArrayList<RemotePoolEvents> _eventListeners;
    private final long _size;
    private String _remoteFile;
    private Status _status;
    private int _downloadProgress;
    private String _downloadError;
    private LocalPool _associatedLocalPool;

    public RemotePool(int elementNr, int format, int version, String name,
                      String dateField, String sum, long size) {
        _elementNr = elementNr;
        _format = format;
        _version = version;
        _name = name;
        _dateField = dateField;
        _sum = sum;
        _size = size;
        _status = Status.rpAvailable;
        _eventListeners = new ArrayList<>();
    }

    public Status getStatus() {
        return _status;
    }

    private void setStatus(Status to) {
        _status = to;
        ArrayList<RemotePoolEvents> handlersCopy = new ArrayList<>(_eventListeners);
        for (RemotePoolEvents handler : handlersCopy) {
            switch (_status) {
                case rpDownloading:
                    handler.onRemotePoolDownloadProgressUpdate(this, _downloadProgress);
                    break;

                case rpDownloadFailed:
                    handler.onRemotePoolDownloadError(this, _downloadError);
                    break;

                case rpDownloaded:
                    handler.onRemotePoolDownloadComplete(this, _associatedLocalPool);
                    break;

                case rpAvailable:
                    // Shouldn't be possible.
                    break;
            }
        }
    }

    public void setDownloading(int progress) {
        _downloadProgress = progress;
        setStatus(Status.rpDownloading);
    }

    public void setDownloadFailed(String message) {
        _downloadError = message;
        setStatus(Status.rpDownloadFailed);
    }

    public void setDownloaded() {
        _downloadProgress = 100;
        setStatus(Status.rpDownloaded);
    }

    public int getDownloadProgress() {
        return _downloadProgress;
    }

    public String getDownloadErrorString() {
        return _downloadError;
    }

    public void addEventHandler(RemotePoolEvents handler) {
        _eventListeners.add(handler);
    }

    public void removeEventHandler(RemotePoolEvents handler) {
        _eventListeners.remove(handler);
    }

    public LocalPool getAssociatedLocalPool() {
        return _associatedLocalPool;
    }

    public void setAssociatedLocalPool(LocalPool pool) {
        _associatedLocalPool = pool;
    }

    public String getRemoteFilename() {
        return _remoteFile;
    }

    public void setRemoteFilename(String fn) {
        _remoteFile = fn;
    }

    public String getSum() {
        return _sum;
    }

    public String getDateField() {
        return _dateField;
    }

    public int getElementNr() {
        return _elementNr;
    }

    public int getFormat() {
        return _format;
    }

    public String getName() {
        return _name;
    }

    public int getVersion() {
        return _version;
    }

    /**
     * \returns The date this pool becomes effective
     */
    public Calendar getStartEffectiveDate() {
        return LocalPool.getEffectiveDateRange(getDateField()).from;
    }

    /**
     * \returns The date this pool stops being effective
     */
    public Calendar getEndEffectiveDate() {
        return LocalPool.getEffectiveDateRange(getDateField()).to;
    }

    /**
     * \returns One of the DATE_STATUS_... constants in the Pool interface,
     * which indicate whether this question pool is active, expired, or a future
     * question pool.
     */
    public int getDateStatus() {
        return LocalPool.getDateStatusFromDatefield(getDateField());
    }

    @SuppressWarnings("unused")
    public long getSize() {
        return _size;
    }

    public enum Status {
        rpAvailable, rpDownloading, rpDownloaded, rpDownloadFailed,
    }
}
