package net.rten.palmve.activities;

import android.app.AlertDialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.os.Bundle;

import androidx.lifecycle.ViewModelProvider;

import net.rten.palmve.LocalPool;
import net.rten.palmve.Preferences;
import net.rten.palmve.R;
import net.rten.palmve.data.pool.Answer;
import net.rten.palmve.data.test.TestState;
import net.rten.palmve.models.QuestionModel;
import net.rten.palmve.models.TestQuestionData;
import net.rten.palmve.models.TestTakingModel;
import net.rten.palmve.repository.PoolRepo;
import net.rten.palmve.repository.TestHistoryRepo;

import java.util.List;

import javax.inject.Inject;

import dagger.hilt.android.AndroidEntryPoint;

@AndroidEntryPoint
public class TestTakingActivity extends QuestionAndAnswersActivity {

    private TestTakingModel _testModel;
    private QuestionModel _questionModel;

    @Inject
    protected PoolRepo _pools;

    @Inject
    protected TestHistoryRepo _testHistory;

    public TestTakingActivity() {
        super(QuestionModel.Mode.answeringQuestion);
    }

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        _testModel = new ViewModelProvider(this).get(TestTakingModel.class);
        _testModel.setRandomizeQuestionOrder(Preferences.getTestTakingRandomOrder(this));
        _testModel.setCorrectAnswersStillWrong(Preferences.getCorrectRetriesStillWrong(this));
        _testModel.getCurrentQuestionIndex().observe(this, index -> {
            List<TestQuestionData> questions = _testModel.getTestQuestions().getValue();
            if (questions != null && index != null) {
                // We have the QuestionView already but we don't do anything with it, the question
                // model will reload it. This could be improved.
                setActiveQuestionId(questions.get(index).poolQuestion.id);
                updateTitle();
            }
        });
        _testModel.getTestProgress().observe(this, progress -> {
            if (progress.percentAnswered == 100) {
                // The test is complete.
                new Thread(_testModel::completeTest).start();
            }
            updateTitle();
        });
        _testModel.getTestInfo().observe(this, info -> {
            if (info.testState != TestState.IN_PROGRESS) {
                Intent intent = new Intent(this, TestResultsActivity.class);
                intent.putExtra(TestResultsActivity.INTENT_POOL_NAME, info.elementName);
                intent.putExtra(TestResultsActivity.INTENT_POOL_DATE, info.elementDateField);
                intent.putExtra(TestResultsActivity.INTENT_TEST_ID, info.id);
                finish();  // Don't stay in the stack.
                startActivity(intent);
            }
        });

        final LocalPool pool = getPool();
        _testModel.setUp(pool.getName(), pool.getDateField(), pool.getRoomDb(), _testHistory.getDb());

        _questionModel = new ViewModelProvider(this).get(QuestionModel.class);
        _questionModel.getSelectedAnswer().observe(this, answer -> new Thread(() -> {
            if (answer == null) {
                // This can happen if we cycle to a new question in the same activity or model.
                //noinspection UnnecessaryReturnStatement
                return;
            } else if (_questionModel.getWrongAnswerBehavior() == Preferences.WrongAnswerBehavior.PROMPT &&
                    _questionModel.getHeaderDataCopy().getValue() != null &&
                    !answer.equals(_questionModel.getHeaderDataCopy().getValue().rightAnswer)) {
                runOnUiThread(() -> {
                    // They got it wrong, but they want to retry.
                    AlertDialog.Builder b = new AlertDialog.Builder(this);
                    DialogInterface.OnClickListener clickListener = (dialog, id) -> {
                        if (id == DialogInterface.BUTTON_NEGATIVE) {
                            new Thread(this::submitAnswer).start();
                        }
                    };

                    b.setMessage(getText(R.string.Browse_Answers_WrongAnswerPrompt));
                    b.setCancelable(false);
                    b.setPositiveButton(R.string.Standard_Yes, clickListener);
                    b.setNegativeButton(R.string.Standard_No, clickListener);
                    b.show();
                });
            } else {
                // Submit the answer and move on.
                submitAnswer();
            }
        }).start());
    }

    private void submitAnswer() {
        // We'll either submit the first selected answer, or the current one, depending on what
        // the preference says for honoring retries.
        Answer.AnswerCode answerToSubmit = _testModel.getCorrectAnswersStillWrong()
                ? _questionModel.getFirstAnswerSelected().getValue()
                : _questionModel.getSelectedAnswer().getValue();
        if (answerToSubmit != null) {
            _testModel.submitAnswerToActiveQuestion(answerToSubmit);
        }
    }

    private void updateTitle() {
        final TestTakingModel testModel = new ViewModelProvider(this).get(TestTakingModel.class);
        List<TestQuestionData> questions = testModel.getTestQuestions().getValue();
        Integer currentQuestion = testModel.getCurrentQuestionIndex().getValue();
        TestTakingModel.TestProgress progress = testModel.getTestProgress().getValue();
        if (questions != null && currentQuestion != null && progress != null) {
            setTitle(getString(R.string.Browse_Answers_TestTitle,
                    questions.get(currentQuestion).testQuestion.questionDesignation,
                    progress.answersProvided + 1, questions.size(),
                    progress.percentAnswered));
        }
    }

    @Override
    protected LocalPool getPool() {
        return _pools.getLocalPoolById(getPoolId());
    }
}
