package net.rten.palmve.asynctasks;

import net.rten.palmve.ChecksumDoesNotMatchException;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Formatter;

public class FetchRemotePoolAsyncTask {
    private static final String DOWNLOAD_EXT_TAG = "-downloading";

    public interface Events {
        void handleProgressUpdates(int readSoFar, int contentLength);
    }

    static public void run(URL url, File downloadTo, String checksum, Events events) throws IOException {
        File downloadFileName = new File(downloadTo.getPath() + DOWNLOAD_EXT_TAG);
        FileOutputStream fileOutputStream = new FileOutputStream(downloadFileName);
        OutputStream outputStream = fileOutputStream;
        MessageDigest digest = null;

        try {
            // if we can, verify the digest of the incoming data
            DigestOutputStream digestOutputStream = new DigestOutputStream(
                    fileOutputStream, MessageDigest.getInstance("MD5"));
            digest = digestOutputStream.getMessageDigest();
            outputStream = digestOutputStream;
        } catch (NoSuchAlgorithmException e) {
            // just ignore it, we won't verify the checksum
        }

        URLConnection connection = url.openConnection();
        int contentLength = connection.getContentLength();
        int readSoFar = 0;
        InputStream inputStream = connection.getInputStream();
        byte[] inputBuffer = new byte[1024];

        // read from the network, write to disk
        while (true) {
            int byteCount = inputStream.read(inputBuffer);
            if (byteCount <= 0) {
                break;
            }
            outputStream.write(inputBuffer, 0, byteCount);
            readSoFar += byteCount;
            events.handleProgressUpdates(readSoFar, contentLength);
        }

        // close the stream and check the digest
        outputStream.close();
        verifyFileSumOrDeleteAndThrow(digest, checksum, downloadFileName);

        // (replace or) create the 'blessed' version of the file
        if (downloadTo.exists()) {
            if (!downloadTo.delete()) {
                throw new RuntimeException("failed to delete " + downloadTo.toString());
            }
        }
        if (!downloadFileName.renameTo(downloadTo)) {
            throw new RuntimeException("failed to rename to " + downloadTo.toString());
        }
    }

    /**
     * Verify that the file we downloaded has the right MD5sum. We could do
     * this with a DigestOutputStream, but it might throw if MD-5 isn't
     * available on the device. I have no idea what's available where, so
     * I've isolated this here to keep from having to muddy up the code.
     *
     * @param digested Container for the message digest. If this is null, no
     *                 validation is performed.
     * @param sum      The expected sum in string format.
     * @throws ChecksumDoesNotMatchException When the sums don't match.
     */
    static private void verifyFileSumOrDeleteAndThrow(
            MessageDigest digested, String sum, File localFile) throws ChecksumDoesNotMatchException {
        if (digested != null) {
            byte[] digest = digested.digest();
            StringBuilder digestString = new StringBuilder();

            try (Formatter formatter = new Formatter(digestString)) {
                for (byte b : digest) {
                    formatter.format("%02x", b);
                }

                // delete the file and raise the exception if the sum doesn't match
                if (!sum.equals(digestString.toString())) {
                    if (!localFile.delete()) {
                        throw new RuntimeException("failed to delete " + localFile.toString());
                    }
                    throw new ChecksumDoesNotMatchException("Checksum validation failed");
                }
            }
        }
    }
}
