package net.rten.palmve.models;

import androidx.lifecycle.LiveData;
import androidx.lifecycle.MutableLiveData;
import androidx.lifecycle.Observer;
import androidx.lifecycle.ViewModel;

import net.rten.palmve.data.pool.PoolDatabase;
import net.rten.palmve.data.test.StatsDao;
import net.rten.palmve.data.test.StatsView;
import net.rten.palmve.data.test.TestCount;
import net.rten.palmve.data.test.TestDatabase;
import net.rten.palmve.data.test.TestState;
import net.rten.palmve.utilities.CopyObserver;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

public class TestHistoryStatsModel extends ViewModel {
    private final CopyStats _copyStats = new CopyStats();
    private final CopyObserver<List<TestCount>> _copyCounts = new CopyObserver<>();
    private final MutableLiveData<Integer> _descriptions = new MutableLiveData<>();
    private LiveData<List<StatsView>> _stats;
    private PoolDatabase _poolDb;
    private TestDatabase _testDb;
    private String _elementName, _elementDate;

    // Default initial settings:
    private StatsDao.DetailReportType _reportType = StatsDao.DetailReportType.questionGroups;
    private boolean _correctAnswers = false;
    private int _secondsAgo = 604800;
    private int _limit = 10;

    public StatsDao.DetailReportType getReportType() {
        return _reportType;
    }

    public boolean forCorrectAnswers() {
        return _correctAnswers;
    }

    public int getSecondsAgo() {
        return _secondsAgo;
    }

    public int getLimit() {
        return _limit;
    }

    public void setReportParams(StatsDao.DetailReportType type, boolean correctAnswers, int secondsAgo, int limit) {
        _reportType = type;
        _correctAnswers = correctAnswers;
        _secondsAgo = secondsAgo;
        _limit = limit;
        resetStatsView();
    }

    public void setUp(String elementName, String elementDate, PoolDatabase poolDb, TestDatabase testDb) {
        if (_poolDb == null) {
            _poolDb = poolDb;
            _testDb = testDb;
            _elementName = elementName;
            _elementDate = elementDate;
            resetStatsView();
        }
    }

    private void resetStatsView() {
        if (_poolDb != null) {
            if (_stats != null) {
                _stats.removeObserver(_copyStats);
            }
            _stats = _testDb.statsDao().getStats(_elementName, _elementDate, _reportType,
                    _correctAnswers, _secondsAgo, _limit);
            _stats.observeForever(_copyStats);

            // test counts aren't going to change while the model is active
            LiveData<List<TestCount>> testCounts;
            if (_secondsAgo > 0) {
                Calendar newerThan = Calendar.getInstance();
                newerThan.add(Calendar.SECOND, -_secondsAgo);
                testCounts = _testDb.testDao().getTestCountsNewerThan(_elementName, _elementDate, newerThan);
            } else {
                testCounts = _testDb.testDao().getTestCounts(_elementName, _elementDate);
            }
            _copyCounts.stopObservingMaster();
            _copyCounts.startObservingMaster(testCounts);
        }
    }

    public LiveData<List<Stats>> getStats() {
        return _copyStats.getCopy();
    }

    public LiveData<List<TestCount>> getCounts() {
        return _copyCounts.getCopy();
    }

    /**
     * Returns an index representing the last item in our stats list whose description text
     * has been loaded from the pool database. It's implemented this way so that you can listen
     * to this item, and then notify a list that the item's current index has changed whenever
     * the observer fires. That will properly wire up a RecyclerView to update the item.
     */
    public LiveData<Integer> getDescriptions() {
        return _descriptions;
    }

    @Override
    public void onCleared() {
        if (_stats != null) {
            _stats.removeObserver(_copyStats);
            _stats = null;
        }
        _copyCounts.stopObservingMaster();
        _poolDb = null;
        _testDb = null;
        _copyStats.clear();
    }

    public int getTestCount() {
        int sum = 0;
        if (_copyCounts.getCopy().getValue() != null) {
            for (TestCount count : _copyCounts.getCopy().getValue()) {
                if (count.testState != TestState.IN_PROGRESS) {
                    sum += count.testCount;
                }
            }
        }
        return sum;
    }

    public int getPassingTestCount() {
        int passingTests = 0;
        if (_copyCounts.getCopy().getValue() != null) {
            for (TestCount count : _copyCounts.getCopy().getValue()) {
                if (count.testState == TestState.PASSED) {
                    passingTests += count.testCount;
                }
            }
        }
        return passingTests;
    }

    public void asyncDeleteTestHistory() {
        new Thread(() -> _testDb.testDao().deleteTestHistory(_elementName, _elementDate)).start();
    }

    // Used to track data during the id -> name resolution process through the pool DB.
    private static class Resolver {
        public LiveData<String> name;
        public Observer<String> observer;
    }

    /**
     * This is the class that our model exports in its list of live data.
     */
    public static class Stats {
        // This comes from the pool database.
        public String questionOrGroupDescription;

        // This comes from the test history database.
        public StatsView testStats;
    }

    private class CopyStats implements Observer<List<StatsView>> {
        final private MutableLiveData<List<Stats>> _copy = new MutableLiveData<>();
        final private ArrayList<Resolver> _resolvers = new ArrayList<>();

        public LiveData<List<Stats>> getCopy() {
            return _copy;
        }

        // Clear out any current resolutions and data.
        public void clear() {
            for (Resolver r : _resolvers) {
                r.name.removeObserver(r.observer);
            }
            _resolvers.clear();
        }

        @Override
        public void onChanged(List<StatsView> item) {
            clear();
            ArrayList<Stats> data = new ArrayList<>();
            for (int i = 0; i < item.size(); ++i) {
                // Make the 'Stats' object which will contain the view and the resolved name.
                final Stats copy = new Stats();
                copy.questionOrGroupDescription = "Loading...";
                copy.testStats = item.get(i);
                data.add(copy);

                // Now make the resolver object, which wraps our query to get the name.
                Resolver r = new Resolver();
                if (_reportType == StatsDao.DetailReportType.questionGroups) {
                    r.name = _poolDb.questionGroupsDao().getQuestionGroupDescription(copy.testStats.questionGroupId);
                } else {
                    r.name = _poolDb.questionDao().getQuestionDescription(copy.testStats.questionId);
                }

                // Now when we finally get the name, we need to set it into the Stats object
                // and post the list back into the live data object.
                final int finalIdx = i;
                r.observer = name -> {
                    copy.questionOrGroupDescription = name;
                    _descriptions.setValue(finalIdx);
                };
                r.name.observeForever(r.observer);
                _resolvers.add(r);
            }
            _copy.setValue(data);
        }
    }
}
